/** @file fwPeriphAddressBACnet.ctl
 * 
 * Functions to set _address config for datapoint for BACnet driver.
 * 
 * Naming: when possible names used in WinCC OA help are used
 * (e.g. "reference" is preferred over "address").
 * FW_PARAMETER_FIELD_AAABBB constants are used, not the fwPeriphAddress_AAABBB
 * (who know which one is correct?).
 * TODO: it would be better to switch to fwPeriphAddress_AAABBB as
 * everybody else is using it.
 *
 * Note that config identifier is "BACnet", but the driver uses "BACNET",
 * in some cases it might cause confusion, that's why there are wrapper
 * functions using "BACNET" have been added. Probably it was better to
 * use "BACNET" from the start even though it's not correct...
 * 
 * Initial code was written based on WinCC OA help file on BACnet
 * and _address config entry. Other things were taken from BACnet
 * examples (BACnet_Samples.dpl in dplist directory of BACnet directory
 * in WinCC OA installation, look for "PeriphAddrMain"). Also
 * panels:
 * - panels/para/address_bacnet.pnl (used to set _address),
 * - panels/para/bacnet.pnl (to create/setup a device; might not be used here).
 * Note: version WinCC OA 3.11 was used.
 * 
 * Table for _datatype. Help only defines 800, but these values were used
 * in various places:
 * 800 default  
 * 801 Boolean  
 * 802 UnsignedInteger  
 * 803 SignedInteger  
 * 804 Real   
 * 805 Double   
 * 809 Enumerated  806 skipped, see bacnetDrvPara.ctl, line 740
 * Valid IDs: 800-899.
 * 
 * Code based on other fwPeriphAddressAAABB.ctl files.
 * 
 * Note that setting _address is not enough to make it work
 * appriopriate BACnet device datapoint has to exist (_BacnetDevice).
 * 
 * Changelog (put only important stuff here):
 * - 2016-01-12: added wrappers for BACNET (lgoralcz & msudera)
 * - 2015-12-08: fixes (lgoralcz)
 * - 2015-11-30: initial creation (lgoralcz)
 * 
 */

const string fwPeriphAddress_TYPE_BACNET = "BACnet"; // Identifier used outside, note that this identifier
                                                     // is used to create function names that are called
                                                     // _get, _set, if you change function names change this
                                                     // well
const string fwPeriphAddress_DRVID_BACNET = "BACNET"; // Driver itself requires it to be all uppercase

// Where are other indexes? See fwPeriphAddress.ctl & FW_PARAMETER_FIELD_AAABBBCCC
const unsigned fwPeriphAddress_BACNET_POLL_GROUP      = 12;
const unsigned fwPeriphAddress_BACNET_OBJECT_SIZE     = 12;

const string fwPeriphAddress_BACNET_REF_SEPARATOR = "."; /// Separator used in reference _address field
const int fwPeriphAddress_BACNET_REF_MIN_ELEM_NO = 4; /// Minimum number of elements in reference _address field
const int fwPeriphAddress_BACNET_REF_MAX_ELEM_NO = 5; /// Maximum number of elements in reference _address field
const int fwPeriphAddress_BACNET_REF_DEV_IDX = 1;  /// Index for device name in peripheral address
const int fwPeriphAddress_BACNET_REF_OBJ_TYPE_IDX = 2; /// Index for object type in peripheral address
const int fwPeriphAddress_BACNET_REF_OBJ_ID_IDX = 3; /// Index for object id in peripheral address
const int fwPeriphAddress_BACNET_REF_PROP_IDX = 4; /// Index for property in peripheral address
const int fwPeriphAddress_BACNET_REF_IDX_IDX = 5; /// Index for property's index in peripheral address

// _datatype possible values are listed here
const int fwPeriphAddress_BACNET_TYPE_DEFAULT = 800; /// Default 
const int fwPeriphAddress_BACNET_TYPE_BOOL = 801; /// Boolean
const int fwPeriphAddress_BACNET_TYPE_UINT = 802; /// Unsigned Integer
const int fwPeriphAddress_BACNET_TYPE_INT = 803; /// Signed Integer
const int fwPeriphAddress_BACNET_TYPE_REAL = 804; /// Real
const int fwPeriphAddress_BACNET_TYPE_DOUBLE = 805; /// Double
const int fwPeriphAddress_BACNET_TYPE_ENUM = 809; /// Enumerated
// In case you need a list of valid _datatypes
const dyn_int fwPeriphAddress_BACNET_TYPES = makeDynInt(
  fwPeriphAddress_BACNET_TYPE_DEFAULT,
  fwPeriphAddress_BACNET_TYPE_BOOL,
  fwPeriphAddress_BACNET_TYPE_UINT,
  fwPeriphAddress_BACNET_TYPE_INT,
  fwPeriphAddress_BACNET_TYPE_REAL,
  fwPeriphAddress_BACNET_TYPE_DOUBLE,
  fwPeriphAddress_BACNET_TYPE_ENUM
);


/** 
 * @brief Internal function to setup the SNMP addressing
 *   
 * @par Constraints
 * Should only be called from fwPeriphAddress_set
 * 
 * @param dpe  Datapoint element to act on
 * @param addressConfig    Address object is passed here:                                                        \n\n
 * 
 * @param exceptionInfo    Details of any errors are returned here
*/
void _fwPeriphAddressBACnet_set(string dpe, dyn_anytype addressConfig, dyn_string &exceptionInfo)
{
  dyn_errClass errors;
  
  // Check if the structure passed is valid (also, correct if necessary)
  _fwPeriphAddressBACnet_check(addressConfig, exceptionInfo);
  if(dynlen(exceptionInfo) > 0)
  {
    return;
  }

  // Set driver number (in case of BACnet there can be only one)
  int driverNum = addressConfig[FW_PARAMETER_FIELD_DRIVER];
  dpSetWait(dpe + ":_distrib.._type", DPCONFIG_DISTRIBUTION_INFO,
            dpe + ":_distrib.._driver", driverNum);
  errors = getLastError();
  if(dynlen(errors) > 0)
  {
    DebugTN(errors);
    throwError(errors);
    fwException_raise(exceptionInfo, "ERROR", "Could not set the driver number.", "");
    return;
  }

  if(addressConfig[FW_PARAMETER_FIELD_LOWLEVEL])
  {
    addressConfig[FW_PARAMETER_FIELD_MODE] += PVSS_ADDRESS_LOWLEVEL_TO_MODE;
  }

  // Set the address config, for reference see scripts/libs/para.ctl
  // function paDpSetAddress(), line 7919. Order, most probably, is important
  dpSetWait(dpe + ":_address.._type", DPCONFIG_PERIPH_ADDR_MAIN,
            dpe + ":_address.._reference", addressConfig[FW_PARAMETER_FIELD_ADDRESS], // BACnet address string
            dpe + ":_address.._mode", (int) addressConfig[FW_PARAMETER_FIELD_MODE], // Explicit conversion to int
                                                                                    // (some internal type mismatch?)
            dpe + ":_address.._datatype", addressConfig[FW_PARAMETER_FIELD_DATATYPE], // 800 - 899, see ETM-1292
            dpe + ":_address.._poll_group", addressConfig[fwPeriphAddress_BACNET_POLL_GROUP],
            dpe + ":_address.._drv_ident", fwPeriphAddress_DRVID_BACNET, // This should always be "BACNET"
            dpe + ":_address.._active", addressConfig[FW_PARAMETER_FIELD_ACTIVE]); // Is this configuration active or not?

  // Any problems?
  errors = getLastError(); 
  if(dynlen(errors) > 0)
  { 
    throwError(errors);
    fwException_raise(exceptionInfo, "ERROR", "Could not create the address config.", "");
  }
  
  return;
}


/** This is to fix problem with different config type identifier ("BACnet") and
 * different driver identifier ("BACNET"). This function is just a wrapper around
 * "proper" function.
 * Proper BACnet name is "BACnet", but for whatever reason driver uses (and some
 * of our panels use this information /fwPeriphAddress.pnl/) "BACNET".
 */
void _fwPeriphAddressBACNET_set(string dpe, dyn_anytype addressConfig, dyn_string &exceptionInfo)
{
  _fwPeriphAddressBACnet_set(dpe, addressConfig, exceptionInfo);
}


/**
 * @brief Retrieve configuration from _address config for BACnet.
 *   
 * @par Constraints
 * Should only be called from fwPeriphAddress_get
 * 
 * @param dpe            Datapoint element to read from
 * @param addressConfig  Address object is returned here (configParameters). @sa _fwPeriphAddressBACnet_set()
 * @param isActive       TRUE if address config is active, else FALSE 
 * @param exceptionInfo  Details of any errors are returned here
*/
void _fwPeriphAddressBACnet_get(string dpe, dyn_anytype &addressConfig, bool &isActive, dyn_string &exceptionInfo)
{
  addressConfig = makeDynAnytype();
  addressConfig[fwPeriphAddress_BACNET_OBJECT_SIZE] = ""; // Force size on list (create it with a given size)
  
  // For description of these fields see either WinCC OA help (see _address) or _fwPeriphAddressBACnet_set()
  dpGet(dpe + ":_distrib.._driver", addressConfig[FW_PARAMETER_FIELD_DRIVER],
        dpe + ":_address.._reference", addressConfig[FW_PARAMETER_FIELD_ADDRESS],
        dpe + ":_address.._mode", addressConfig[FW_PARAMETER_FIELD_MODE],
        dpe + ":_address.._datatype", addressConfig[FW_PARAMETER_FIELD_DATATYPE],
        dpe + ":_address.._poll_group", addressConfig[fwPeriphAddress_BACNET_POLL_GROUP],
        dpe + ":_address.._active", addressConfig[FW_PARAMETER_FIELD_ACTIVE]);

  // Any problems?
  dyn_errClass errors = getLastError(); 
  if(dynlen(errors) > 0)
  { 
    throwError(errors);
    fwException_raise(exceptionInfo, "ERROR", "Could not get address config for " + dpe + ".", "");
    return;
  }
  
  // Detect low level
  if(addressConfig[FW_PARAMETER_FIELD_MODE] >= PVSS_ADDRESS_LOWLEVEL_TO_MODE)
  {
    addressConfig[FW_PARAMETER_FIELD_LOWLEVEL] = true;
  }
  
  // Set driver identification manually, we are using "BACnet", but the driver uses "BACNET"
  addressConfig[FW_PARAMETER_FIELD_COMMUNICATION] = fwPeriphAddress_TYPE_BACNET;
  
  isActive = addressConfig[FW_PARAMETER_FIELD_ACTIVE];
}


/** This is to fix problem with different config type identifier ("BACnet") and
 * different driver identifier ("BACNET"). This function is just a wrapper around
 * "proper" function.
 * Proper BACnet name is "BACnet", but for whatever reason driver uses (and some
 * of our panels use this information /fwPeriphAddress.pnl/) "BACNET".
 */
void _fwPeriphAddressBACNET_get(string dpe, dyn_anytype &addressConfig, bool &isActive, dyn_string &exceptionInfo)
{
  _fwPeriphAddressBACnet_get(dpe, addressConfig, isActive, exceptionInfo);
}


/**
 * @brief Internal function to cleanup before deleting the BACnet _address and _distrib configs
 * TODO: create function body
 * @par Constraints
 * Should only be called from fwPeriphAddress_delete \n
 * 
 *
 * @param dpe            Datapoint element to read from
 * @param exceptionInfo  Details of any errors are returned here 
 */ 
_fwPeriphAddressBACnet_delete(string dpe, dyn_string &exceptionInfo)
{
  DebugTN("_fwPeriphAddressBACnet_delete(): not implemented");
  
  return;
}


/** 
 * @brief Function to check and correct BACnet address configuration parameters before
 *        attempting to save them to the DP element.
 * 
 * Note: in other cases this function is public (no "_" prefix), but there's no reason
 *       why to make it private.
 *
 * @param addressConfig The address configuration object is passed here. In some cases, an
 *                      amended (fixed) version may be returned here
 * @param exceptionInfo Details of errors in the address configuration are returned here
*/
void _fwPeriphAddressBACnet_check(dyn_anytype &addressConfig, dyn_string &exceptionInfo)
{
  // Check size...
  if(dynlen(addressConfig) != fwPeriphAddress_BACNET_OBJECT_SIZE)
  {
    fwException_raise(exceptionInfo, "WARNING", "_fwPeriphAddressBACnet_check: " +
                                                "bad size of config structure", "");

    // Return, config is not reliable (possible problems with out-of-range)
    return;
  }

  // Is this a BACnet config?
  if(addressConfig[FW_PARAMETER_FIELD_COMMUNICATION] != fwPeriphAddress_TYPE_BACNET)
  {
    fwException_raise(exceptionInfo, "ERROR", "_fwPeriphAddressBACnet_check: " +
                                              "this seems not to be a BACnet config", "");
  }
  
  // Check _datatype
  if(!_fwPeriphAddressBACnet_isDataTypeValid(addressConfig[FW_PARAMETER_FIELD_DATATYPE]))
  {
    fwException_raise(exceptionInfo, "ERROR", "_fwPeriphAddressBACnet_check: " +
                                              "not a valid _datatype (not 800-899)", "");
  }

  // Check reference field
  if(!_fwPeriphAddressBACnet_isReferenceValid(addressConfig[FW_PARAMETER_FIELD_ADDRESS]))
  {
    fwException_raise(exceptionInfo, "ERROR", "_fwPeriphAddressBACnet_check: " +
                                              "reference address is invalid", "");
  }

  // TODO: chcek with standard BACnet address set panel
  string pollGroup = addressConfig[fwPeriphAddress_BACNET_POLL_GROUP];
  
  switch(addressConfig[FW_PARAMETER_FIELD_MODE])
  {
    case DPATTR_ADDR_MODE_INPUT_POLL:
    case DPATTR_ADDR_MODE_IO_POLL:
      if(strlen(pollGroup) == 0)
      {
        fwException_raise(exceptionInfo, "ERROR", "_fwPeriphAddressBACnet_check: " +
                                                  "A poll group must be defined for polled input configurations", "");
      }
      else
      {
        //if poll group is defined, does not being with a "_" and does not specify the system name, then prepend the "_"
        if(strpos(pollGroup,"_")!=0 && strpos(pollGroup,":")<0)
          pollGroup = "_"+pollGroup;
    
        if(!dpExists(pollGroup))
          fwException_raise(exceptionInfo, "ERROR", "_fwPeriphAddressBACnet_check: " +
                                                    "The poll group \"" + pollGroup + "\" does not exist", "");
      }      
      break;
    case DPATTR_ADDR_MODE_INPUT_SPONT:
    case DPATTR_ADDR_MODE_INPUT_SQUERY:
    case DPATTR_ADDR_MODE_OUTPUT:
    case DPATTR_ADDR_MODE_IO_SQUERY:
    case DPATTR_ADDR_MODE_IO_SPONT:
      pollGroup = ""; //clear poll group because it is not needed
      break;
    default:
      fwException_raise(exceptionInfo, "ERROR", "_fwPeriphAddressBACnet_check: " +
                                                "Mode " + addressConfig[FW_PARAMETER_FIELD_MODE] + " is not supported", "");
      break;
  }
  // Update poll group, in some cases it might be corrected
  addressConfig[fwPeriphAddress_BACNET_POLL_GROUP] = pollGroup;

  // TODO: anything else?
}


/** This is to fix problem with different config type identifier ("BACnet") and
 * different driver identifier ("BACNET"). This function is just a wrapper around
 * "proper" function.
 * Proper BACnet name is "BACnet", but for whatever reason driver uses (and some
 * of our panels use this information /fwPeriphAddress.pnl/) "BACNET".
 */
void _fwPeriphAddressBACNET_check(dyn_anytype &addressConfig, dyn_string &exceptionInfo)
{
  _fwPeriphAddressBACnet_check(addressConfig, exceptionInfo);
}


/** Validate datatype.
 * 
 * @param dataType	(string)	IN datatype (as string), ex. "800". 
 * @return value of type 'bool' true if datatype is acceptable, false if not
 */
 bool _fwPeriphAddressBACnet_isDataTypeValid(string dataType)
 {
   int dt = (int) dataType; // Convert to int
   
   return dynContains(fwPeriphAddress_BACNET_TYPES, dt) > 0;
 }


/** Validate BACnet address reference. 
 * 
 * Example reference: "Device_6789.AnalogInput.1.Min_Pres_Value"
 * or: "Device_6789.BinaryValue.3.Reliability.2".
 * Note: last element is optional.
 * 
 * <Device>.<Object_Type>.<Object_Id>.<Property>.<Index /optional/>
 * 
 * @param reference	(string)	IN BACnet address reference
 * @return value of type 'bool' true if address is valid, false otherwise
 */
bool _fwPeriphAddressBACnet_isReferenceValid(string reference)
{
  dyn_string elements = _fwPeriphAddressBACnet_explodeReference(reference);
  int elemCount = dynlen(elements);
  if(elemCount == 0)
  {
    return false;
  }
  
  string device = elements[fwPeriphAddress_BACNET_REF_DEV_IDX];
  string objectType = elements[fwPeriphAddress_BACNET_REF_OBJ_TYPE_IDX];
  string objectId = elements[fwPeriphAddress_BACNET_REF_OBJ_ID_IDX];
  string property = elements[fwPeriphAddress_BACNET_REF_PROP_IDX];
  string index = "";
  // This one is optional
  if(elemCount > fwPeriphAddress_BACNET_REF_MIN_ELEM_NO)
  {
    index = elements[fwPeriphAddress_BACNET_REF_IDX_IDX];
  }
  
  // Now check each element
  bool isValid = _fwPeriphAddressBACnet_isDeviceValid(device) &&
                 _fwPeriphAddressBACnet_isObjecTypeValid(objectType) &&
                 _fwPeriphAddressBACnet_isObjectIdValid(objectId) &&
                 _fwPeriphAddressBACnet_isPropertyValid(objectType, property);
  if(index != "")
  {
    isValid = isValid && _fwPeriphAddressBACnet_isIndexValid(index);
  }

  return isValid;
}


/** Explodes string representing BACnet reference. * 
 * 
 * @param reference	(string)	IN BACnet reference string, ex. " 
 * @return value of type 'dyn_string' List of elements (ex. BACnet device, object id, etc.) or
 *                                    empty list if string was invalid
 * 
 */
dyn_string _fwPeriphAddressBACnet_explodeReference(string reference)
{
  // Quote from WinCC OA help on BACnet
  // <Device>.<Object_Type>.<Object_Id>.<Property>.<Index>
  // From examples last element is optional
  dyn_string elements = strsplit(reference, fwPeriphAddress_BACNET_REF_SEPARATOR);
  int elemCount = dynlen(elements);
  
  if((elemCount > fwPeriphAddress_BACNET_REF_MAX_ELEM_NO) ||
     (elemCount < fwPeriphAddress_BACNET_REF_MIN_ELEM_NO))
  {
    // Invalid number of elements, explicitly clear output list
    DebugTN("Invalid BACnet reference string given, number of elements (" +
            elemCount + ") not in range (" + fwPeriphAddress_BACNET_REF_MIN_ELEM_NO +
            " - " + fwPeriphAddress_BACNET_REF_MAX_ELEM_NO + ").");
    dynClear(elements);
  }
  
  return elements;
}


/** Check if a given BACnet device exists in the configuration.
 * 
 * This function simply check if a datapoint of a given name
 * and type "_BacnetDevice" exists. 
 * 
 * TODO: stub, returns always true
 * 
 * @param device	(string)	IN BACnet device name
 * @return value of type 'bool' True if device exits, false if BACnet device doesn't exist
 */
bool _fwPeriphAddressBACnet_isDeviceValid(string device)
{
  return true;
}

bool _fwPeriphAddressBACnet_isObjecTypeValid(string objectType)
{
  // Taken from catalog file BACnetGeneral.cat
  dyn_string validObjTypes = makeDynString(
    "AnalogInput",
    "AnalogOutput",
    "AnalogValue",
    "BinaryInput",
    "BinaryOutput",
    "BinaryValue",
    "Calendar",
    "Device",
    "MultistateInput",
    "MultistateOutput",
    "NotificationClass",
    "Schedule",
    "MultistateValue",
    "TrendLog",
    "Command",
    "EventEnrolment",
    "File",
    "Group",
    "Loop",
    "Program",
    "Averaging",
    "LifeSafetyPoint",
    "LifeSafetyZone",
    "Accumulator",
    "PulseConverter");

  return dynContains(validObjTypes, objectType) > 0;
}


/** Check if BACnet object ID is valid for _reference element.
 * 
 * Note, this does not check if an object id really exists in
 * the device, only if it is correct, i.e. a number. 
 * 
 * @param objectId	(string)	IN BACnet object id
 * @return value of type 'bool' true if valid BACnet object id, false if not valid
 */
bool _fwPeriphAddressBACnet_isObjectIdValid(string objectId)
{
  return true;
}


/** Check if BACnet object's property is valid for _reference element. 
 *
 * Note1: that cat file that is being used is not the same cat file that
 * is used to store translated messages ("msg" directory).
 * Note2: code based on one of the WinCC OA BACnet libraries.
 * 
 * @param property	(string)	IN BACnet object's property
 * @return value of type 'bool' true if property is valid, false otherwise
 */
bool _fwPeriphAddressBACnet_isPropertyValid(string object, string property)
{
  string sContent, sFilePath, sType;
  file fFile;
  dyn_string dsRows;
  int err;
  
  sFilePath = getPath(DATA_REL_PATH,"bacnet/BACNet_objecttype_mapping.cat");
  fFile = fopen(sFilePath,"r");

  if((err=ferror(fFile)) != 0)
  {
    DebugN("Error "+err+" opening file "+ sFilePath + ". Warning: Property cannot be validated." +
	       " _fwPeriphAddressBACnet_isPropertyValid() return true");
    return true;
  }

  fileToString(sFilePath, sContent);
  dsRows = strsplit(sContent, "\n");  
  for (int i = 1; i <= dynlen(dsRows); i++)
  {
	if(dsRows[i][0] != "-" && dsRows[i] != "")
	{
	  dyn_string dsRow = strsplit(dsRows[i], ",");
	  if(dsRow[1] == object)
	  {
        i++;
        while((dynlen(dsRows) >= i) && (dsRows[i][0] == "-"))
        {
          if(strltrim(dsRows[i], "-") == property)
		  {
			fclose(fFile);
		    return true;
		  }
          i++;
        }
		break;
      }
    }
  }
  fclose(fFile);
  
  return false;
}


/** Check if given BACnet object's property index is valid.
 * 
 * @param index	(string)	IN object's property index
 * @return value of type 'bool' true if BACnet object's property index is valid, false otherwise
 */
bool _fwPeriphAddressBACnet_isIndexValid(string index)
{
  return true;
}
