/**
* This file contains OPCUA specific functions for fwPeriphAddress.
*/

/**
* Get a list of subscriptions on the system for the given server.
* 
* @param sys            The system.
* @param server         The server.
* @param &exceptionInfo Exceptions.
*
* @return dyn_string A list of subscriptions.
*/
public dyn_string fwPeriphAddressOPCUA_getSubscriptions(string sys, string server, dyn_string &exceptionInfo) {

  dyn_string dsSub;
  
  if (strlen(sys) == 0) {
     fwException_raise(exceptionInfo, "ERROR", "No system","");
     return makeDynString();
  }  
  
  string dpe = sys + "_" +  server + ".Config.Subscriptions:_original.._value";

  if (!dpExists(dpe)) {
    if (strlen(server) == 0) {
       fwException_raise(exceptionInfo, "ERROR", "No OPCUA server","");
       return dsSub;
    }
    fwException_raise(exceptionInfo, "ERROR", "Subscriptions DPE does not exist","");
    return dsSub;
  }  
  
  int retVal = dpGet(dpe, dsSub);
  
  if (retVal == -1) {
    fwException_raise(exceptionInfo, "ERROR", "Could not get subscriptions: dpGet(" + dpe + ") failed","");
    return dsSub;
  }  
  
  int len = dynlen(dsSub);
  
  for (int i = 1; i <= len; i++) {
    dsSub[i] = dpSubStr(dsSub[i], DPSUB_DP);
    dsSub[i] = substr(dsSub[i], 1, strlen(dsSub[i]) - 1);
  }
  
  dynSortAsc(dsSub);

  return dsSub;
}

/**
* Get the type for the given subscription.
*
* @param sys          The system.
* @param subscription The subscription.
*
* @return int The type or -1 on error.
*/
public int fwPeriphAddressOPCUA_getSubscriptionType(string sys, string subscription, dyn_string &exceptionInfo) {
  int type;

  if (strlen(sys) == 0) {
     fwException_raise(exceptionInfo, "ERROR", "No system","");
     return -1;
  }  
  
  string dpe = sys + "_" + subscription + ".Config.SubscriptionType";

  if (!dpExists(dpe)) {
    if (strlen(subscription) == 0) {
       fwException_raise(exceptionInfo, "ERROR", "No subscription","");
       return -1;
    }
    fwException_raise(exceptionInfo, "ERROR", "Subscription type DPE does not exist","");
    return -1;
  }      
  
  int retVal = dpGet(dpe, type);
  
  if (retVal == -1) {
    fwException_raise(exceptionInfo, "ERROR", "Could not get subscription type: dpGet(" + dpe + ") failed","");
    return -1;
  }  
  
  return type;
}

/**
* Get a list of servers on the given system.
*
* @return dyn_string List of servers.
*/
public dyn_string fwPeriphAddressOPCUA_getServers(string sys, dyn_string &exceptionInfo) {
  if (strlen(sys) == 0) {
     fwException_raise(exceptionInfo, "ERROR", "No system","");
     return makeDynString();
  }
  
  dyn_string dsEqu = dpNames(sys+"*", "_OPCUAServer");
  
  int len = dynlen(dsEqu);
  
  for (int i = len; i > 0; i--) {
    // don't display redundant datapoints
    if (isReduDp(dsEqu[i])) {
      dynRemove(dsEqu, i);
    }
  }
  len = dynlen(dsEqu);
  if (len > 0) {
    for (int i = 1; i <= len; i++) {
      dsEqu[i] = dpSubStr(dsEqu[i], DPSUB_DP);
      dsEqu[i] = substr(dsEqu[i], 1, strlen(dsEqu[i]) - 1);
    }
  }

  return dsEqu;
}
